;Allows to detect current code generation mode (2,4,8)
macro detectMode byteCount
{
    virtual
        mov eax,[0]
        byteCount = 1 shl ($-$$-3)
    end virtual
}

;Displays a binary representation of a number
macro dispBin num*, padding, leader, trailer, size
{
    local digCount,number,lastdig
    number = size num
    lastdig = number

    if number < 0
        display '-'
        number = -(number shr 1 + number and 1)
        lastdig = -(lastdig + number + number)
    else
        number = number shr 1
        lastdig = lastdig and 1
    end if
    digCount = 0
    while number shr digCount > 0
        digCount = digCount + 1
    end while
    if ~ leader eq
        display leader
    end if
    if ~ padding eq
        if digCount < padding
            times (padding-digCount-1) display '0'
        end if
    end if
    times digCount display number shr (digCount-%) and 1+'0'
    display lastdig+'0'
    if ~ trailer eq
        display trailer
    end if
}

;Displays a decimal representation of a number
macro dispDec num*, padding, leader, trailer
{
    local digCount,tenPow,number,lastdig
    number = num
    lastdig = number
    
    if number < 0
        display '-'
        number = (-(number shr 1 + number and 1)) / 5
        lastdig = -(lastdig + number*5 + number*5)
    else
        number = number/10
        lastdig = lastdig mod 10
    end if
    digCount = 0
    tenPow = 1
    while tenPow <= number
        tenPow = tenPow*10
        digCount = digCount + 1
    end while
    if ~ leader eq
        display leader
    end if
    if ~ padding eq
        if digCount < padding
            times (padding-digCount-1) display '0'
        end if
    end if
    repeat digCount
        tenPow = tenPow/10
        display number/tenPow+'0'
        number = number mod tenPow
    end repeat
    display lastdig+'0'
    if ~ trailer eq
        display trailer
    end if
}

;Displays a hexadecimal representation of a number
macro dispHex num*, padding, leader, trailer, size
{
    local digCount,dig,number,lastdig
    number = size num
    lastdig = number
    
    if number < 0
        display '-'
        number = (-(number shr 1 + number and 1)) shr 3
        lastdig = -(lastdig + number shl 3 + number shl 3)
    else
        number = number shr 4
        lastdig = lastdig and 0xF
    end if
    digCount = 0
    while number shr (digCount*4) > 0
        digCount = digCount + 1
    end while
    if ~ leader eq
        display leader
    end if
    if ~ padding eq
        if digCount < padding
            times (padding-digCount-1) display '0'
        end if
    end if
    repeat digCount
        dig = number shr ((digCount-%)*4) and 0Fh
        if dig > 9
            dig = dig-'0'+('A'-10)
        end if
        display dig+'0'
    end repeat
    if lastdig > 9
        lastdig = lastdig-'0'+('A'-10)
    end if
    display lastdig+'0'
    if ~ trailer eq
        display trailer
    end if
}

;Displays a string with binary, decimal and hexadecimal values
;usage: xdisplay d=number,":",9,b=number,9,h=number,13,10
macro xdisplay [token]
{
    forward
        define matched +
        define matched -
        irp type, =b, =d, =h, =bb, =bw, =bd, =bq, =hb, =hw, =hd, =hq \{ match type==n,token \\{ restore matched \\} \}
        match =b==number,token \{ dispBin number,,,'b' \}
        match =d==number,token \{ dispDec number \}
        match =h==number,token \{ dispHex number,,'0x' \}
        match =bb==number,token \{ dispBin number,,,'b',byte \}
        match =bw==number,token \{ dispBin number,,,'b',word \}
        match =bd==number,token \{ dispBin number,,,'b',dword \}
        match =bq==number,token \{ dispBin number,,,'b',qword \}
        match =hb==number,token \{ dispHex number,,'0x',,byte \}
        match =hw==number,token \{ dispHex number,,'0x',,word \}
        match =hd==number,token \{ dispHex number,,'0x',,dword \}
        match =hq==number,token \{ dispHex number,,'0x',,qword \}
        match -,matched \{ display token \}
        match -,matched \{ restore matched \}
        restore matched
}

;Allows using structures as macros (i.e. a structure will be unnamed)
;usage: defStrucsAsMacros POINT,LINE
;       POINT
;       LINE <1,2>,<1,4>
macro defStrucsAsMacros [name*]
{
    forward
        macro name [arg]
        \{
            \common
                \local ..lbl
                ..lbl name arg
        \}
}

;Allows to save a string onto the stack
;usage: pushstr "This is gonna be a stack string",0
struc pushstr [arg*]
{
    common local pushCount, buf, modeByteCount
    
    detectMode modeByteCount
    virtual
        db arg
        pushCount = ($-$$+modeByteCount-1)/modeByteCount
    end virtual
    
    . = pushCount*modeByteCount
    
    repeat pushCount
        virtual
            db arg
            db (modeByteCount-1) dup 0
            if modeByteCount = 2
                load buf word from $$+(pushCount-%)*modeByteCount
            else if modeByteCount = 4
                load buf dword from $$+(pushCount-%)*modeByteCount
            else if modeByteCount = 8
                if % < pushCount
                    load buf qword from $$+(pushCount-%-1)*modeByteCount
                else
                    load buf qword from $$+(pushCount-1)*modeByteCount
                end if
            else
                display 'Error: unknown code generation mode',13,10
                err
            end if
        end virtual
        if modeByteCount = 8
            push rax
            mov rax,buf
        else
            push buf
        end if
    end repeat
    if modeByteCount = 8
        xchg rax,qword[rsp+(pushCount-1)*modeByteCount]
    end if
}
struc pushstrU [arg*]
{
    common local pushCount, buf, modeByteCount
    
    detectMode modeByteCount
    virtual
        du arg
        pushCount = ($-$$+modeByteCount-1)/modeByteCount
    end virtual
    
    . = pushCount*modeByteCount
    
    repeat pushCount
        virtual
            du arg
            db (modeByteCount-1) dup 0
            if modeByteCount = 2
                load buf word from $$+(pushCount-%)*modeByteCount
            else if modeByteCount = 4
                load buf dword from $$+(pushCount-%)*modeByteCount
            else if modeByteCount = 8
                if % < pushCount
                    load buf qword from $$+(pushCount-%-1)*modeByteCount
                else
                    load buf qword from $$+(pushCount-1)*modeByteCount
                end if
            else
                display 'Error: unknown code generation mode',13,10
                err
            end if
        end virtual
        if modeByteCount = 8
            push rax
            mov rax,buf
        else
            push buf
        end if
    end repeat
    if modeByteCount = 8
        xchg rax,qword[rsp+(pushCount-1)*modeByteCount]
    end if
}
defStrucsAsMacros pushstr, pushstrU


iset equ bt,in,ja,jb,jc,je,jg,jl,\
jo,jp,js,jz,or,aaa,aad,aam,\
aas,adc,add,and,bsf,bsr,btc,btr,\
bts,cbw,cdq,clc,cld,cli,cmc,cmp,\
cqo,cwd,daa,das,dec,div,fld,fst,\
hlt,inc,ins,int,jae,jbe,jge,jle,\
jmp,jna,jnb,jnc,jne,jng,jnl,jno,\
jnp,jns,jnz,jpe,jpo,lar,lds,lea,\
les,lfs,lgs,lsl,lss,ltr,mov,mul,\
neg,nop,not,out,pop,por,rcl,rcr,\
rep,ret,rol,ror,rsm,sal,sar,sbb,\
shl,shr,stc,std,sti,str,sub,ud2,\
xor,adcx,adox,andn,arpl,blci,blcs,blsi,\
blsr,bzhi,call,cdqe,clac,clgi,clts,cmps,\
cwde,dppd,dpps,emms,fabs,fadd,fbld,fchs,\
fcom,fcos,fdiv,feni,fild,fist,fld1,fldz,\
fmul,fnop,fsin,fstp,fsub,ftst,fxam,fxch,\
idiv,imul,insb,insd,insw,int1,int3,into,\
invd,iret,jcxz,jnae,jnbe,jnge,jnle,lahf,\
lgdt,lidt,lldt,lmsw,lock,lods,loop,movd,\
movq,movs,mulx,orpd,orps,outs,pand,pdep,\
pext,popa,popd,popf,popq,popw,push,pxor,\
repe,repz,retd,retf,retn,retq,retw,rorx,\
sahf,salc,sarx,scas,seta,setb,setc,sete,\
setg,setl,seto,setp,sets,setz,sgdt,shld,\
shlx,shrd,shrx,sidt,sldt,smsw,stac,stgi,\
stos,test,verr,verw,vpor,wait,xadd,xchg,\
xend,xlat,addpd,addps,addsd,addss,andpd,andps,\
bextr,blcic,blsic,bound,bswap,cmova,cmovb,cmovc,\
cmove,cmovg,cmovl,cmovo,cmovp,cmovs,cmovz,cmppd,\
cmpps,cmpsb,cmpsd,cmpsq,cmpss,cmpsw,cpuid,crc32,\
divpd,divps,divsd,divss,enter,extrq,f2xm1,faddp,\
fbstp,fclex,fcomi,fcomp,fdisi,fdivp,fdivr,femms,\
ffree,fiadd,ficom,fidiv,fimul,finit,fistp,fisub,\
fldcw,fldpi,fmulp,fneni,fprem,fptan,fsave,fsqrt,\
fstcw,fstsw,fsubp,fsubr,fucom,fwait,fyl2x,icebp,\
iretd,iretq,iretw,jecxz,jrcxz,lddqu,leave,lodsb,\
lodsd,lodsq,lodsw,loopd,loope,loopq,loopw,loopz,\
lzcnt,maxpd,maxps,maxsd,maxss,minpd,minps,minsd,\
minss,movbe,movsb,movsd,movsq,movss,movsw,movsx,\
movzx,mulpd,mulps,mulsd,mulss,mwait,outsb,outsd,\
outsw,pabsb,pabsd,pabsw,paddb,paddd,paddq,paddw,\
pandn,pause,pavgb,pavgw,pf2id,pf2iw,pfacc,pfadd,\
pfmax,pfmin,pfmul,pfrcp,pfsub,pi2fd,pi2fw,popad,\
popaw,popfd,popfq,popfw,pslld,psllq,psllw,psrad,\
psraw,psrld,psrlq,psrlw,psubb,psubd,psubq,psubw,\
ptest,pusha,pushd,pushf,pushq,pushw,rcpps,rcpss,\
rdmsr,rdpmc,rdtsc,repne,repnz,retfd,retfq,retfw,\
retnd,retnq,retnw,scasb,scasd,scasq,scasw,setae,\
setbe,setge,setle,setna,setnb,setnc,setne,setng,\
setnl,setno,setnp,setns,setnz,setpe,setpo,stosb,\
stosd,stosq,stosw,subpd,subps,subsd,subss,tzcnt,\
tzmsk,vdppd,vdpps,vmovd,vmovq,vmrun,vmxon,vorpd,\
vorps,vpand,vpxor,wrmsr,xlatb,xorpd,xorps,xsave,\
xtest,aesdec,aesenc,aesimc,andnpd,andnps,blcmsk,blsmsk,\
cmovae,cmovbe,cmovge,cmovle,cmovna,cmovnb,cmovnc,cmovne,\
cmovng,cmovnl,cmovno,cmovnp,cmovns,cmovnz,cmovpe,cmovpo,\
comisd,comiss,fcmovb,fcmove,fcmovu,fcomip,fcompp,fdivrp,\
ffreep,ficomp,fidivr,fisttp,fisubr,fldenv,fldl2e,fldl2t,\
fldlg2,fldln2,fnclex,fndisi,fninit,fnsave,fnstcw,fnstsw,\
fpatan,fprem1,frstor,frstpm,fsaved,fsavew,fscale,fsetpm,\
fstenv,fsubrp,fucomi,fucomp,fxsave,getsec,haddpd,haddps,\
hsubpd,hsubps,invept,invlpg,lfence,llwpcb,looped,loopeq,\
loopew,loopne,loopnz,loopzd,loopzq,loopzw,lwpins,lwpval,\
mfence,movapd,movaps,movdqa,movdqu,movhpd,movhps,movlpd,\
movlps,movnti,movntq,movsxd,movupd,movups,paddsb,paddsw,\
pextrb,pextrd,pextrq,pextrw,pfnacc,pfsubr,phaddd,phaddw,\
phsubd,phsubw,pinsrb,pinsrd,pinsrq,pinsrw,pmaxsb,pmaxsd,\
pmaxsw,pmaxub,pmaxud,pmaxuw,pminsb,pminsd,pminsw,pminub,\
pminud,pminuw,pmuldq,pmulhw,pmulld,pmullw,popcnt,psadbw,\
pshufb,pshufd,pshufw,psignb,psignd,psignw,pslldq,psrldq,\
psubsb,psubsw,pswapd,pushad,pushaw,pushfd,pushfq,pushfw,\
rdmsrq,rdrand,rdseed,rdtscp,setalc,setnae,setnbe,setnge,\
setnle,sfence,shufpd,shufps,skinit,slwpcb,sqrtpd,sqrtps,\
sqrtsd,sqrtss,swapgs,sysret,t1mskc,vaddpd,vaddps,vaddsd,\
vaddss,vandpd,vandps,vcmppd,vcmpps,vcmpsd,vcmpss,vdivpd,\
vdivps,vdivsd,vdivss,vlddqu,vmaxpd,vmaxps,vmaxsd,vmaxss,\
vmcall,vminpd,vminps,vminsd,vminss,vmload,vmovsd,vmovss,\
vmread,vmsave,vmulpd,vmulps,vmulsd,vmulss,vmxoff,vpabsb,\
vpabsd,vpabsw,vpaddb,vpaddd,vpaddq,vpaddw,vpandn,vpavgb,\
vpavgw,vpcmov,vpcomb,vpcomd,vpcomq,vpcomw,vpermd,vpermq,\
vpperm,vprotb,vprotd,vprotq,vprotw,vpshab,vpshad,vpshaq,\
vpshaw,vpshlb,vpshld,vpshlq,vpshlw,vpslld,vpsllq,vpsllw,\
vpsrad,vpsraw,vpsrld,vpsrlq,vpsrlw,vpsubb,vpsubd,vpsubq,\
vpsubw,vptest,vrcpps,vrcpss,vsubpd,vsubps,vsubsd,vsubss,\
vxorpd,vxorps,wbinvd,wrmsrq,xabort,xbegin,xgetbv,xrstor,\
xsetbv,blcfill,blendpd,blendps,blsfill,clflush,cmovnae,cmovnbe,\
cmovnge,cmovnle,cmpeqpd,cmpeqps,cmpeqsd,cmpeqss,cmplepd,cmpleps,\
cmplesd,cmpless,cmpltpd,cmpltps,cmpltsd,cmpltss,cmpxchg,fcmovbe,\
fcmovnb,fcmovne,fcmovnu,fdecstp,fincstp,fldenvd,fldenvw,fnsaved,\
fnsavew,fnstenv,frndint,frstord,frstorw,fsincos,fstenvd,fstenvw,\
fucomip,fucompp,fxrstor,fxtract,fyl2xp1,insertq,invlpga,invpcid,\
invvpid,ldmxcsr,loopned,loopneq,loopnew,loopnzd,loopnzq,loopnzw,\
monitor,movddup,movdq2q,movhlps,movlhps,movntdq,movntpd,movntps,\
movntsd,movntss,movq2dq,mpsadbw,paddusb,paddusw,palignr,pavgusb,\
pblendw,pcmpeqb,pcmpeqd,pcmpeqq,pcmpeqw,pcmpgtb,pcmpgtd,pcmpgtq,\
pcmpgtw,pfcmpeq,pfcmpge,pfcmpgt,pfpnacc,pfrsqrt,phaddsw,phsubsw,\
pmaddwd,pmulhrw,pmulhuw,pmuludq,pshufhw,pshuflw,psubusb,psubusw,\
roundpd,roundps,roundsd,roundss,rsqrtps,rsqrtss,stmxcsr,syscall,\
sysexit,sysretq,ucomisd,ucomiss,vaesdec,vaesenc,vaesimc,vandnpd,\
vandnps,vcomisd,vcomiss,vfrczpd,vfrczps,vfrczsd,vfrczss,vhaddpd,\
vhaddps,vhsubpd,vhsubps,vmclear,vmmcall,vmovapd,vmovaps,vmovdqa,\
vmovdqu,vmovhpd,vmovhps,vmovlpd,vmovlps,vmovupd,vmovups,vmptrld,\
vmptrst,vmwrite,vpaddsb,vpaddsw,vpcomub,vpcomud,vpcomuq,vpcomuw,\
vpermpd,vpermps,vpextrb,vpextrd,vpextrq,vpextrw,vphaddd,vphaddw,\
vphsubd,vphsubw,vpinsrb,vpinsrd,vpinsrq,vpinsrw,vpmaxsb,vpmaxsd,\
vpmaxsw,vpmaxub,vpmaxud,vpmaxuw,vpminsb,vpminsd,vpminsw,vpminub,\
vpminud,vpminuw,vpmuldq,vpmulhw,vpmulld,vpmullw,vpsadbw,vpshufb,\
vpshufd,vpsignb,vpsignd,vpsignw,vpslldq,vpsllvd,vpsllvq,vpsravd,\
vpsrldq,vpsrlvd,vpsrlvq,vpsubsb,vpsubsw,vshufpd,vshufps,vsqrtpd,\
vsqrtps,vsqrtsd,vsqrtss,vtestpd,vtestps,xsave64,addsubpd,addsubps,\
blendvpd,blendvps,cmpneqpd,cmpneqps,cmpneqsd,cmpneqss,cmpnlepd,cmpnleps,\
cmpnlesd,cmpnless,cmpnltpd,cmpnltps,cmpnltsd,cmpnltss,cmpordpd,cmpordps,\
cmpordsd,cmpordss,cvtdq2pd,cvtdq2ps,cvtpd2dq,cvtpd2pi,cvtpd2ps,cvtpi2pd,\
cvtpi2ps,cvtps2dq,cvtps2pd,cvtps2pi,cvtsd2si,cvtsd2ss,cvtsi2sd,cvtsi2ss,\
cvtss2sd,cvtss2si,fcmovnbe,fnstenvd,fnstenvw,fxsave64,insertps,maskmovq,\
movmskpd,movmskps,movntdqa,movshdup,movsldup,packssdw,packsswb,packusdw,\
packuswb,pblendvb,pfrcpit1,pfrcpit2,pfrsqit1,pmovmskb,pmovsxbd,pmovsxbq,\
pmovsxbw,pmovsxdq,pmovsxwd,pmovsxwq,pmovzxbd,pmovzxbq,pmovzxbw,pmovzxdq,\
pmovzxwd,pmovzxwq,pmulhrsw,prefetch,rdfsbase,rdgsbase,sysenter,sysexitq,\
unpckhpd,unpckhps,unpcklpd,unpcklps,vblendpd,vblendps,vcmpeqpd,vcmpeqps,\
vcmpeqsd,vcmpeqss,vcmpgepd,vcmpgeps,vcmpgesd,vcmpgess,vcmpgtpd,vcmpgtps,\
vcmpgtsd,vcmpgtss,vcmplepd,vcmpleps,vcmplesd,vcmpless,vcmpltpd,vcmpltps,\
vcmpltsd,vcmpltss,vfmaddpd,vfmaddps,vfmaddsd,vfmaddss,vfmsubpd,vfmsubps,\
vfmsubsd,vfmsubss,vldmxcsr,vmlaunch,vmovddup,vmovhlps,vmovlhps,vmovntdq,\
vmovntpd,vmovntps,vmpsadbw,vmresume,vpaddusb,vpaddusw,vpalignr,vpblendd,\
vpblendw,vpcmpeqb,vpcmpeqd,vpcmpeqq,vpcmpeqw,vpcmpgtb,vpcmpgtd,vpcmpgtq,\
vpcmpgtw,vpcomeqb,vpcomeqd,vpcomeqq,vpcomeqw,vpcomgeb,vpcomged,vpcomgeq,\
vpcomgew,vpcomgtb,vpcomgtd,vpcomgtq,vpcomgtw,vpcomleb,vpcomled,vpcomleq,\
vpcomlew,vpcomltb,vpcomltd,vpcomltq,vpcomltw,vphaddbd,vphaddbq,vphaddbw,\
vphadddq,vphaddsw,vphaddwd,vphaddwq,vphsubbw,vphsubdq,vphsubsw,vphsubwd,\
vpmacsdd,vpmacswd,vpmacsww,vpmaddwd,vpmulhuw,vpmuludq,vpshufhw,vpshuflw,\
vpsubusb,vpsubusw,vroundpd,vroundps,vroundsd,vroundss,vrsqrtps,vrsqrtss,\
vstmxcsr,vucomisd,vucomiss,vzeroall,wrfsbase,wrgsbase,xacquire,xrelease,\
xrstor64,xsaveopt,cmpxchg8b,cvttpd2dq,cvttpd2pi,cvttps2dq,cvttps2pi,cvttsd2si,\
cvttss2si,extractps,fxrstor64,pclmulqdq,pcmpestri,pcmpestrm,pcmpistri,pcmpistrm,\
pmaddubsw,prefetchw,punpckhbw,punpckhdq,punpckhwd,punpcklbw,punpckldq,punpcklwd,\
vaddsubpd,vaddsubps,vblendvpd,vblendvps,vcmpneqpd,vcmpneqps,vcmpneqsd,vcmpneqss,\
vcmpngepd,vcmpngeps,vcmpngesd,vcmpngess,vcmpngtpd,vcmpngtps,vcmpngtsd,vcmpngtss,\
vcmpnlepd,vcmpnleps,vcmpnlesd,vcmpnless,vcmpnltpd,vcmpnltps,vcmpnltsd,vcmpnltss,\
vcmpordpd,vcmpordps,vcmpordsd,vcmpordss,vcvtdq2pd,vcvtdq2ps,vcvtpd2dq,vcvtpd2ps,\
vcvtph2ps,vcvtps2dq,vcvtps2pd,vcvtps2ph,vcvtsd2si,vcvtsd2ss,vcvtsi2sd,vcvtsi2ss,\
vcvtss2sd,vcvtss2si,vfnmaddpd,vfnmaddps,vfnmaddsd,vfnmaddss,vfnmsubpd,vfnmsubps,\
vfnmsubsd,vfnmsubss,vinsertps,vmovmskpd,vmovmskps,vmovntdqa,vmovshdup,vmovsldup,\
vpackssdw,vpacksswb,vpackusdw,vpackuswb,vpblendvb,vpcomequb,vpcomequd,vpcomequq,\
vpcomequw,vpcomgeub,vpcomgeud,vpcomgeuq,vpcomgeuw,vpcomgtub,vpcomgtud,vpcomgtuq,\
vpcomgtuw,vpcomleub,vpcomleud,vpcomleuq,vpcomleuw,vpcomltub,vpcomltud,vpcomltuq,\
vpcomltuw,vpcomneqb,vpcomneqd,vpcomneqq,vpcomneqw,vpermilpd,vpermilps,vphaddubd,\
vphaddubq,vphaddubw,vphaddudq,vphadduwd,vphadduwq,vpmacsdqh,vpmacsdql,vpmacssdd,\
vpmacsswd,vpmacssww,vpmadcswd,vpmovmskb,vpmovsxbd,vpmovsxbq,vpmovsxbw,vpmovsxdq,\
vpmovsxwd,vpmovsxwq,vpmovzxbd,vpmovzxbq,vpmovzxbw,vpmovzxdq,vpmovzxwd,vpmovzxwq,\
vpmulhrsw,vunpckhpd,vunpckhps,vunpcklpd,vunpcklps,aesdeclast,aesenclast,cmpunordpd,\
cmpunordps,cmpunordsd,cmpunordss,cmpxchg16b,loadall286,loadall386,maskmovdqu,phminposuw,\
prefetcht0,prefetcht1,prefetcht2,punpckhqdq,punpcklqdq,vcmptruepd,vcmptrueps,vcmptruesd,\
vcmptruess,vcvttpd2dq,vcvttps2dq,vcvttsd2si,vcvttss2si,vextractps,vgatherdpd,vgatherdps,\
vgatherqpd,vgatherqps,vmaskmovpd,vmaskmovps,vpclmulqdq,vpcmpestri,vpcmpestrm,vpcmpistri,\
vpcmpistrm,vpcomnequb,vpcomnequd,vpcomnequq,vpcomnequw,vpcomtrueb,vpcomtrued,vpcomtrueq,\
vpcomtruew,vperm2f128,vperm2i128,vpermil2pd,vpermil2ps,vpgatherdd,vpgatherdq,vpgatherqd,\
vpgatherqq,vpmacssdqh,vpmacssdql,vpmadcsswd,vpmaddubsw,vpmaskmovd,vpmaskmovq,vpunpckhbw,\
vpunpckhdq,vpunpckhwd,vpunpcklbw,vpunpckldq,vpunpcklwd,vzeroupper,xsaveopt64,pclmulhqhdq,\
pclmullqhdq,prefetchnta,vaesdeclast,vaesenclast,vcmpeq_ospd,vcmpeq_osps,vcmpeq_ossd,vcmpeq_osss,\
vcmpeq_uqpd,vcmpeq_uqps,vcmpeq_uqsd,vcmpeq_uqss,vcmpeq_uspd,vcmpeq_usps,vcmpeq_ussd,vcmpeq_usss,\
vcmpfalsepd,vcmpfalseps,vcmpfalsesd,vcmpfalsess,vcmpge_oqpd,vcmpge_oqps,vcmpge_oqsd,vcmpge_oqss,\
vcmpgt_oqpd,vcmpgt_oqps,vcmpgt_oqsd,vcmpgt_oqss,vcmple_oqpd,vcmple_oqps,vcmple_oqsd,vcmple_oqss,\
vcmplt_oqpd,vcmplt_oqps,vcmplt_oqsd,vcmplt_oqss,vcmpord_spd,vcmpord_sps,vcmpord_ssd,vcmpord_sss,\
vcmpunordpd,vcmpunordps,vcmpunordsd,vcmpunordss,vfmadd132pd,vfmadd132ps,vfmadd132sd,vfmadd132ss,\
vfmadd213pd,vfmadd213ps,vfmadd213sd,vfmadd213ss,vfmadd231pd,vfmadd231ps,vfmadd231sd,vfmadd231ss,\
vfmaddsubpd,vfmaddsubps,vfmsub132pd,vfmsub132ps,vfmsub132sd,vfmsub132ss,vfmsub213pd,vfmsub213ps,\
vfmsub213sd,vfmsub213ss,vfmsub231pd,vfmsub231ps,vfmsub231sd,vfmsub231ss,vfmsubaddpd,vfmsubaddps,\
vinsertf128,vinserti128,vmaskmovdqu,vpcomfalseb,vpcomfalsed,vpcomfalseq,vpcomfalsew,vpcomtrueub,\
vpcomtrueud,vpcomtrueuq,vpcomtrueuw,vphminposuw,vpunpckhqdq,vpunpcklqdq,pclmulhqhqdq,pclmulhqlqdq,\
pclmullqhqdq,pclmullqlqdq,vbroadcastsd,vbroadcastss,vcmpneq_oqpd,vcmpneq_oqps,vcmpneq_oqsd,vcmpneq_oqss,\
vcmpneq_ospd,vcmpneq_osps,vcmpneq_ossd,vcmpneq_osss,vcmpneq_uspd,vcmpneq_usps,vcmpneq_ussd,vcmpneq_usss,\
vcmpnge_uqpd,vcmpnge_uqps,vcmpnge_uqsd,vcmpnge_uqss,vcmpngt_uqpd,vcmpngt_uqps,vcmpngt_uqsd,vcmpngt_uqss,\
vcmpnle_uqpd,vcmpnle_uqps,vcmpnle_uqsd,vcmpnle_uqss,vcmpnlt_uqpd,vcmpnlt_uqps,vcmpnlt_uqsd,vcmpnlt_uqss,\
vextractf128,vextracti128,vfnmadd132pd,vfnmadd132ps,vfnmadd132sd,vfnmadd132ss,vfnmadd213pd,vfnmadd213ps,\
vfnmadd213sd,vfnmadd213ss,vfnmadd231pd,vfnmadd231ps,vfnmadd231sd,vfnmadd231ss,vfnmsub132pd,vfnmsub132ps,\
vfnmsub132sd,vfnmsub132ss,vfnmsub213pd,vfnmsub213ps,vfnmsub213sd,vfnmsub213ss,vfnmsub231pd,vfnmsub231ps,\
vfnmsub231sd,vfnmsub231ss,vpbroadcastb,vpbroadcastd,vpbroadcastq,vpbroadcastw,vpclmulhqhdq,vpclmullqhdq,\
vpcomfalseub,vpcomfalseud,vpcomfalseuq,vpcomfalseuw,vpermilmo2pd,vpermilmo2ps,vpermilmz2pd,vpermilmz2ps,\
vpermiltd2pd,vpermiltd2ps,vcmptrue_uspd,vcmptrue_usps,vcmptrue_ussd,vcmptrue_usss,vcmpunord_spd,vcmpunord_sps,\
vcmpunord_ssd,vcmpunord_sss,vpclmulhqlqdq,vpclmullqlqdq,vbroadcastf128,vbroadcasti128,vcmpfalse_ospd,vcmpfalse_osps,\
vcmpfalse_ossd,vcmpfalse_osss,vfmaddsub132pd,vfmaddsub132ps,vfmaddsub213pd,vfmaddsub213ps,vfmaddsub231pd,vfmaddsub231ps,\
vfmsubadd132pd,vfmsubadd132ps,vfmsubadd213pd,vfmsubadd213ps,vfmsubadd231pd,vfmsubadd231ps,aeskeygenassist,vaeskeygenassist

match ,
{
    local iset_s, totalSize_s, maxSize_s, maxSizeVar_s
    macro ilen_ [iset]
    \{
        \common \local totalSize, maxSize, maxSizeVar, inmacro, omit
        define totalSize_s totalSize
        define maxSize_s maxSize
        define maxSizeVar_s maxSizeVar
        iset_s equ iset
        
        totalSize = 0
        maxSizeVar = 0
        match \\iset,iset \\{ irp i,\\iset
        \\\{
            macro i [args]
            \\\\{
                \\\\common \\\\local iAddr, iSize, buf
                    inmacro = 1
                    iAddr = $
                    i args
                    match =omit,omit
                    \\\\\{
                        if inmacro ;inmacro might have become 0
                            iSize = $-iAddr
                            totalSize = totalSize+iSize
                            if iSize > maxSizeVar
                                maxSizeVar = iSize
                            end if
                            
                            if iAddr relativeto 0
                                dispHex iAddr,,'0x'
                            else
                                dispHex rva iAddr,,'0x'
                            end if
                            display ': '
                            times -((iSize/10)-1)shr 64 display ' '
                            dispDec iSize
                            display ' | '
                            repeat iSize
                                load buf byte from iAddr+(%-1)
                                dispHex buf,2,,' '
                            end repeat
                            times maxSize-iSize display '   '
                            display '| ',\\\\\`i
                            
                            define _token +
                            irps arg,args
                            \\\\\\{
                                \\\\\\common display ' '
                                \\\\\\forward
                                    match \\\\\\`arg,arg \\\\\\\{ rept 0 { \\\\\\\}
                                    match l \\\\\\`arg t,'' '+' '-' '*' '/' '=' '(' ')' '[' ']' '{' '}' ':' ',' '|' '&' '~' ''
                                    \\\\\\\{
                                        define token +
                                        rept 0 {
                                    \\\\\\\}
                                    define token -
                                    match -,_token \\\\\\\{ display ' ' \\\\\\\}
                                    restore _token
                                    _token equ token
                                    restore token
                                    match \\\\\\`arg,arg \\\\\\\{ display "'" \\\\\\\}
                                    display \\\\\\`arg
                                    match \\\\\\`arg,arg \\\\\\\{ display "'" \\\\\\\}
                            \\\\\\}
                            restore _token
                            display 13,10
                        end if
                    \\\\\}
                    inmacro = 0
            \\\\}
        \\\} \\}
        macro virtual [arg]
        \\{
            \\common virtual arg
            define omit +
        \\}
        struc virtual [arg]
        \\{
            \\common . virtual arg
            match =end,. \\\{ restore omit \\\}
        \\}
        macro ilen_ [arg]
        \\{
            \\common
                display 'Error: _ilen missing',13,10
                err
        \\}
    \}
    macro _ilen
    \{
        match iset,iset_s \\{ purge iset \\}
        purge virtual, ilen_
        restruc virtual

        match =totalSize_s,totalSize_s
        \\{
            display "Error: ilen_ missing",13,10
            err
        \\}
        maxSize_s = maxSizeVar_s
        xdisplay 'Total size of the shown instructions: ',d=totalSize_s,13,10
        restore iset_s, totalSize_s, maxSize_s
    \}
}

match ,
{
    local begin_s, delim_s, mload_s
    macro disphex_ delim
    \{
        \local begin, mload, buf, omit
        
        define begin_s begin
        define mload_s mload
        define delim_s delim
        begin = $

        macro mload \\{\\}
        macro org [arg] \\{ \\common match =omit,omit
        \\\{
            \\\local ..space, sbegin, send
            sbegin = begin
            send = $
            ..space::

            org arg
            begin = $

            macro mload
            \\\\{
                mload
                purge mload
                repeat send-sbegin
                    load buf byte from ..space:sbegin+(%-1)
                    db buf
                end repeat
            \\\\}
        \\\} match +,omit \\\{ org arg \\\} \\}
        macro virtual [arg]
        \\{
            \\common virtual arg
            define omit +
        \\}
        struc virtual [arg]
        \\{
            \\common . virtual arg
            match =end,. \\\{ restore omit \\\}
        \\}
        macro disphex_ [arg]
        \\{
            \\common
                display 'Error: _disphex missing',13,10
                err
        \\}
    \}
    macro _disphex
    \{
        \local dig,buf,..space,ssize
        
        match =begin_s,begin_s
        \\{
            display 13,10,"Error: disphex_ missing",13,10
            err
        \\}
        
        ssize = $-begin_s
        ..space::
        virtual at 0
            match mload,mload_s
            \\{
                mload
                purge mload
            \\}
            repeat ssize
                load buf byte from ..space:begin_s+(%-1)
                db buf
            end repeat

            repeat $-$$
                load buf byte from %-1
                repeat 2
                    dig = buf shr ((2-%)*4) and 0Fh+'0'
                    if dig > '9'
                        dig = dig-'0'+('A'-10)
                    end if
                    display dig
                end repeat
                if % < $
                    match any,delim_s \\{ display delim_s \\}
                end if
            end repeat
        end virtual
        
        purge org,virtual
        restruc virtual
        restore begin_s, mload_s, delim_s
        purge disphex_
    \}
}

match ,
{
    local ..hextbl, begin_s, size_s, delim_s, mload_s, mstore_s
    
    virtual at 0
        ..hextbl::
        ;irp x,0 1 2 3 4 5 6 7 8 9 A B C D E F{irps i,x{irps j,x{db\`i,\\`j\\}\}}
        db 16 dup '000102030405060708090A0B0C0D0E0F'
        times 10*16 store byte (%-1)/16+'0' at (%-1)*2
        times 6*16  store byte (%-1)/16+'A' at (%-1+10*16)*2
    end virtual
    
    macro enchex_ delim
    \{
        \local begin, size, mload, mstore, buf, omit
        
        define begin_s begin
        define size_s size
        define delim_s delim
        define mload_s mload
        define mstore_s mstore
        
        begin = $
        size = 0
        
        macro mload \\{\\}
        macro mstore src \\{\\}
        macro org [arg] \\{ \\common match =omit,omit
        \\\{
            \\\local ..space, sbegin, send
            sbegin = begin
            send = $
            size = size + (send-sbegin)
            ..space::
            
            org arg
            begin = $
            
            macro mload
            \\\\{
                mload
                purge mload
                repeat send-sbegin
                    load buf byte from ..space:sbegin+(%-1)
                    db buf
                end repeat
            \\\\}
            macro mstore src
            \\\\{
                mstore src
                purge mstore
                repeat send-sbegin
                    load buf byte from src+(%-1)
                    store byte buf at ..space:sbegin+(%-1)
                end repeat
                src = src + (send-sbegin)
            \\\\}
        \\\} match +,omit \\\{ org arg \\\} \\}
        macro virtual [arg]
        \\{
            \\common virtual arg
            define omit +
        \\}
        struc virtual [arg]
        \\{
            \\common . virtual arg
            match =end,. \\\{ restore omit \\\}
        \\}
    \}
    
    macro _enchex
    \{
        \local hexohead, sizeof.delim, i, buf, ..space, ssize
        
        match =begin_s,begin_s
        \\{
            display "Error: enchex_ missing",13,10
            err
        \\}
        if delim_s eq
            sizeof.delim = 0
        else
            virtual
                db delim_s
                sizeof.delim = $-$$
            end virtual
        end if
        
        ssize = $-begin_s
        size_s = size_s + ssize
        hexohead = size_s*(sizeof.delim+1)
        if hexohead > 0
            hexohead = hexohead - sizeof.delim
        end if
        rb hexohead
        ..space::
        
        virtual at 0
            rb hexohead
            match mload,mload_s
            \\{
                mload
                purge mload
            \\}
            repeat ssize
                load buf byte from ..space:begin_s+(%-1)
                db buf
            end repeat
            
            repeat size_s
                i = %-1
                ;store byte hexcode
                load buf byte from hexohead+i
                load buf word from ..hextbl:buf*2
                store word buf at i*(2+sizeof.delim)
                if % < size_s
                    ;store delim
                    repeat sizeof.delim
                        virtual
                            db delim_s
                            load buf byte from $$+(%-1)
                        end virtual
                        store byte buf at i*(2+sizeof.delim)+(%-1)+2
                    end repeat
                end if
            end repeat
            i = $$
            match mstore,mstore_s
            \\{
                mstore i
                purge mstore
            \\}
            repeat $-i
                load buf byte from i+(%-1)
                store byte buf at ..space:begin_s+(%-1)
            end repeat
        end virtual
        
        purge org,virtual
        restruc virtual
        restore begin_s, size_s, delim_s, mload_s, mstore_s
    \}
}

match ,
{
    local begin_s, base_ns, mload_s, base_s, awidth_ns, np_s
    
    macro dispmem_ base, awidth=8, np='.' ;display base, address width, nonprintable characters
    \{
        \local begin,base_n,awidth_n,mload,buf,omit
        
        define begin_s begin
        define base_ns base_n
        define awidth_ns awidth_n
        define mload_s mload
        define base_s base
        define np_s np

        begin = $
        match any, base \\{ base_n = base \\}
        awidth_n = awidth

        macro mload \\{\\}
        macro org [arg] \\{ \\common match =omit,omit
        \\\{
            \\\local ..space, sbegin, send
            sbegin = begin
            send = $
            ..space::

            org arg
            begin = $

            macro mload
            \\\\{
                mload
                purge mload
                match , base \\\\\{ base_n = sbegin \\\\\}
                dq base_n,send-sbegin
                repeat send-sbegin
                    load buf byte from ..space:sbegin+(%-1)
                    db buf
                end repeat
                match any, base \\\\\{ base_n = base_n+(send-sbegin) \\\\\}
            \\\\}
        \\\} match +,omit \\\{ org arg \\\} \\}
        macro virtual [arg]
        \\{
            \\common virtual arg
            define omit +
        \\}
        struc virtual [arg]
        \\{
            \\common . virtual arg
            match =end,. \\\{ restore omit \\\}
        \\}
        macro dispmem_ [arg]
        \\{
            \\common
                display 'Error: _dispmem missing',13,10
                err
        \\}
    \}
    
    macro _dispmem
    \{
        \local dig,buf,curptr,displen,b,e,..space,ssize,spaceptr
        match =np_s,np_s
        \\{
            display "Error: dispmem_ missing",13,10
            err
        \\}
        
        ssize = $-begin_s
        ..space::
        virtual at 0
            match mload,mload_s
            \\{
                mload
                purge mload
            \\}
            match , base_s \\{ base_ns = begin_s \\}
            dq base_ns, ssize
            repeat ssize
                load buf byte from ..space:begin_s+(%-1)
                db buf
            end repeat
            
            spaceptr = $$
            while spaceptr < $
                load b qword from spaceptr
                load e qword from spaceptr+8
                if b >= 1 shl 63
                    b = b - 1 shl 63 - 1 shl 63
                end if
                e = b+e
                spaceptr = spaceptr+16
                
                if e-b > 0
                    times awidth_ns display ' '
                    display '   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F',13,10
                    
                    curptr = b and not $F
                    displen = ((e+$F) and not $F - curptr) shr 4
                    repeat displen
                        ;display address
                        repeat awidth_ns
                            dig = curptr shr ((awidth_ns-%)*4) and 0Fh+'0'
                            if dig > '9'
                                dig = dig-'0'+('A'-10)
                            end if
                            display dig
                        end repeat
                        display ': '
                        
                        ;display hex data
                        repeat $10
                            if curptr+(%-1) < b | e <= curptr+(%-1)
                                display '   '
                            else
                                load buf byte from spaceptr+(%-1)+(curptr-b)
                                repeat 2
                                    dig = buf shr ((2-%)*4) and 0Fh+'0'
                                    if dig > '9'
                                        dig = dig-'0'+('A'-10)
                                    end if
                                    display dig
                                end repeat
                                display ' '
                            end if
                        end repeat
                        display ' |'
                        
                        ;display ascii data
                        repeat $10
                            if curptr+(%-1) < b | e <= curptr+(%-1)
                                display ' '
                            else
                                load buf byte from spaceptr+(%-1)+(curptr-b)
                                if buf < $20
                                    match np,np_s \\{ display \\`np \\}
                                else
                                    display buf
                                end if
                            end if
                        end repeat
                        display '|',13,10
                        curptr = curptr + $10
                    end repeat
                end if
                spaceptr = spaceptr+(e-b)
            end while
        end virtual
        
        purge org,virtual
        restruc virtual
        restore begin_s, base_ns, mload_s, base_s, awidth_ns, np_s
        purge dispmem_
    \}
}
